#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/keysym.h>
#include <X11/Xaw/Simple.h>

#include "dftype.h"
#include "dfxfunc.h"

#define FONT_SIZE 14

static int CursorUp(Widget w, DfFileList *list);
static int CursorDown(Widget w, DfFileList *list);
static int PageUp(Widget w, DfFileList *list);
static int PageDown(Widget w, DfFileList *list);
static void DoOpen(Widget w, DfFileList *list);

static int ChDir(DfFileList *list, const char *new_dir);
static void SelectFile(Widget w, DfFileList *list);
static int GotoParent(Widget w, DfFileList *list);

static void wndDrawProc(DfWindow *list, Display *display, Widget w, GC gc);

static void wndResize(Widget w, XtPointer p, XEvent *event, Boolean *cont);
static void wndKeyProc(Widget w, XtPointer p, XEvent *event, Boolean *cont);
static void wndExpose(Widget w, XtPointer p, XEvent *event, Boolean *cont);

static void CalcCursorRect(DfFileList *list, int nPos, XRectangle *rc);
extern int ShowDialog(Widget w);

static int DoCopy(Widget w, DfFileList  *list)
{
  DfStr s;
  int n;
  int i;
  char *p;

  n = MakeSelectList(&s, list);
  p = Str_Get(&s);
  printf("%d files select.\n", list->nSelects);
  for(i = 0; i < n; i++){
    printf("%d:%s\n", i + 1, p);
    p += (strlen(p) + 1);
  }

  Str_Free(&s, 0);

  ShowDialog(w);

  return 0;
}

/* will be use XtGrabKeyboard() */
static void wndKeyProc(Widget w, XtPointer p, XEvent *event, Boolean *cont)
{
  XKeyEvent *ev;
  KeySym k;
  DfWindow *dfx;
  DfFileList *list;
  int redraw;
  int pre_cur;
  XRectangle rc;

  redraw = 0;
  dfx = p;
  list = dfx->lists;
  pre_cur = list->nCur;

  ev = (XKeyEvent*)event;
  k = XLookupKeysym(ev, 0);
  switch(k){
  case XK_a:
    break;
  case XK_c:
    DoCopy(w, list);
    break;
  case XK_q:
    exit(0);
    break;
  case XK_x:
    DumpMemInfo();
    break;
  case XK_Up:
    redraw = CursorUp(w, list);
    break;
  case XK_Down:
    redraw = CursorDown(w, list);
    break;
  case XK_Return:
    DoOpen(w, list);
    break;
  case XK_space:
    SelectFile(w, list);
    redraw = CursorDown(w, list);
    if(!redraw && pre_cur == list->nCur){
      CalcCursorRect(list, list->nCur, &rc);
      XClearArea(XtDisplay(w), XtWindow(w), rc.x, rc.y, rc.width, rc.height, True);
    }
    break;
  case XK_Prior:
    redraw = PageUp(w, list);
    break;
  case XK_Next:
    redraw = PageDown(w, list);
    break;
  case XK_BackSpace:
    redraw = GotoParent(w, list);
    break;
  }

  if(redraw){
    InvalidateWindow(w);
    return;
  }
  if(pre_cur != list->nCur){
    CalcCursorRect(list, pre_cur, &rc);
    XClearArea(XtDisplay(w), XtWindow(w), rc.x, rc.y, rc.width, rc.height, True);
    CalcCursorRect(list, list->nCur, &rc);
    XClearArea(XtDisplay(w), XtWindow(w), rc.x, rc.y, rc.width, rc.height, True);
  }
}

static int CursorUp(Widget w, DfFileList *list)
{
  int n;

  n = list->nCur;
  if(n){
    n--;
    list->nCur = n;
    if(n < list->nStart){
      list->nStart--;
      return 1;
    }
  }
 
  return 0;
}

static int CursorDown(Widget w, DfFileList *list)
{
  int n;

  n = list->nCur;
  if(n < list->nFiles - 1){
    n++;
    list->nCur = n;
    if(list->nStart + list->rc.height / vnFontHeight - 1 < n){
      /* scroll */
      list->nStart++;
      //     XCopyArea(XtDisplay(w), XtWindow(w), XtWindow(w), DefaultGC(XtDisplay(w), 0), 0, vnFontHeight, list->width, list->height, 0, 0);
      return 1;
    }
  }

  return 0;
}

static int PageUp(Widget w, DfFileList *list)
{
  int n;
  int pre;

  pre = list->nCur;

  n = list->rc.height / vnFontHeight;
  switch(n){/* :D  */
  case 0:
    break;
  default:
    n--;
  }
  if(list->nCur <= n){
    list->nCur = 0;
  }else{
    list->nCur -= n;
  }
  if(list->nStart <= n){
    list->nStart = 0;
  }else{
    list->nStart -= n;
  }

  if(pre != list->nCur){
    return 1;
  }
  return 0;
}

static int PageDown(Widget w, DfFileList *list)
{
  int n;
  int pre;

  pre = list->nCur;

  n = list->rc.height / vnFontHeight;
  switch(n){/* :D  */
  case 0:
  case 1:
    break;
  default:
    n--;
  }
  list->nCur += n;
  if(list->nFiles <= list->nCur){
    list->nCur = list->nFiles - 1;
  }

  if(list->nStart + n < list->nCur){
    list->nStart += n;
    if(list->nFiles - n - 1 <= list->nStart){
      list->nStart = list->nFiles - n - 1;
    }
  }

  if(pre != list->nCur){
    return 1;
  }
  return 0;
}


static void wndExpose(Widget w, XtPointer p, XEvent *event, Boolean *cont)
{
  XExposeEvent *ev;
  DfWindow *dfx;
  GC gc;

  ev = (XExposeEvent*)event;

  dfx = p;
  if(ev->count){
    /* add to clip mask and exit. */
    if(dfx->fClipping){
      AddRect(&dfx->clip, ev->x, ev->y, ev->width, ev->height);
    }else{
      dfx->fClipping = 1;
      dfx->clip.x = ev->x;
      dfx->clip.y = ev->y;
      dfx->clip.width = ev->width;
      dfx->clip.height = ev->height;
    }
    return;
  }

  /* drawing */
  dfx->clip.x = ev->x;
  dfx->clip.y = ev->y;
  dfx->clip.width = ev->width;
  dfx->clip.height = ev->height;

  gc = DefaultGC(ev->display, DefaultScreen(ev->display));

  wndDrawProc(dfx, ev->display, w, gc);
}

static void wndDrawProc(DfWindow *dfx, Display *display, Widget w, GC gc)
{
  DfFile *files;
  DfFile *cur;
  DfFileList *list;
  int x;
  int y;
  int cx;
  int cy;
  int n;
  int nResidual;
  long int fg;
  long int bg;
  XRectangle rc;

  XSetClipRectangles(display, gc, 0, 0, &dfx->clip, 1, Unsorted);

  /* ToDo:
   * multi file lists.
   */
  list = dfx->lists;
  cx = list->rc.width;
  cy = vnFontHeight;
  files = list->files + list->nStart;
  nResidual = list->nFiles - list->nStart;
  cur = list->files + list->nCur;
  for(x = list->rc.x; x < list->rc.width; x += cx){
    n = 0;
    for(y = vnFontBase; y - vnFontBase < list->rc.height; y += cy){
      fg = files->color;
      bg = colors[COLOR_BK];
      if(files == cur){
	fg = colors[COLOR_CURFG];
	bg = colors[COLOR_CURBG];
      }else if(files->nSelect){
	fg = colors[COLOR_SELFG];
	bg = colors[COLOR_SELBG];
      }
      CalcCursorRect(list, n + list->nStart, &rc);
      XSetForeground(display, gc, bg);
      XFillRectangle(XtDisplay(w), XtWindow(w), gc, rc.x, rc.y, rc.width, rc.height);

      if(files->nSelect){
	XSetForeground(display, gc, colors[COLOR_MARK]);
	XFillRectangle(XtDisplay(w), XtWindow(w), gc, rc.x, rc.y, rc.x + 4, rc.height);
      }
      XSetForeground(display, gc, fg);
      XmbDrawString(display, XtWindow(w), vFontSet, gc, x + 4, y, files->pszName, strlen(files->pszName));

      files++;
      nResidual--;
      if(nResidual == 0){
	goto DRAW_FINISH;
      }
      n++;
    }
  }

 DRAW_FINISH:
  dfx->fClipping = 0;
}

static void wndResize(Widget w, XtPointer p, XEvent *event, Boolean *cont){
  XConfigureEvent *ev;
  DfWindow *dfx;
  DfFileList *list;
  int cur;
  int n;

  switch(event->type){
  case ConfigureNotify:
    dfx = (DfWindow*)p;
    ev = (XConfigureEvent*)event;

    dfx->width = ev->width;
    dfx->height = ev->height;

    /* temporary */
    list = dfx->lists;
    list->rc.width = dfx->width;
    list->rc.height = dfx->height;

    cur = list->nCur;
    if(list->rc.height < vnFontHeight){
      list->nStart = cur;
    }else{
      n = list->rc.height / vnFontHeight;
      if(n){
        n--;
      }
      if(list->nStart + n < cur){
        if(cur - n <= 0){
          n = 0;
        }else{
          n = cur - n;
        }
        list->nStart = n;
      }
    }
    break;
  }
}



/* filer window */
void Wnd_Create(Widget parent, int cx, int cy)
{
  Widget w;
  DfWindow *dfx;
  DfFileList *list;
  char buf[256];

  w = XtVaCreateManagedWidget("main", simpleWidgetClass, parent,
			      XtNwidth, cx,
			      XtNheight, cy,
			      XtNborderWidth, 4,
			      NULL);
  dfx = bMalloc(sizeof(*dfx));
  list = bMalloc(sizeof(*list));
  memset(dfx, 0, sizeof(*dfx));
  memset(list, 0, sizeof(*list));
  dfx->w = w;
  dfx->lists = list;
  dfx->width = cx;
  dfx->height = cy;

  list->nCur = 0;
  list->nStart = 0;
  list->pszCurDir = NULL;
  list->rc.x = 0;
  list->rc.y = 0;
  list->rc.width = cx;
  list->rc.height = cy;
  Wnd_Initialize(w, dfx);
  getcwd(buf, sizeof(buf));
  GetFiles(list, buf);
}

void Wnd_Initialize(Widget w, DfWindow *dfx)
{
  /* initialize */
  /* add event handler */
  XtAddEventHandler(w, KeyPressMask, False, wndKeyProc, dfx);
  XtAddEventHandler(w, ExposureMask, False, wndExpose, dfx);
  XtAddEventHandler(w, StructureNotifyMask, False, wndResize, dfx);
}

static void DoOpen(Widget w, DfFileList *list)
{
  int cur;
  int attr;

  cur = list->nCur;
  attr = list->files[cur].attr;

  if(S_ISDIR(attr)){
    if(ChDir(list, list->files[cur].pszName)){
      InvalidateWindow(w);
    }
    return;
  }
}

static int ChDir(DfFileList *list, const char *new_dir)
{
  printf("new: %s\n", new_dir);
  if(GetFiles(list, new_dir)){
    list->nCur = 0;
    list->nStart = 0;
    return 1;
  }

  return 0;
}

static void CalcCursorRect(DfFileList *list, int nPos, XRectangle *rc)
{
  nPos -= list->nStart;
  rc->x = list->rc.x;
  rc->y = list->rc.y + nPos * vnFontHeight;
  rc->width = list->rc.width;
  rc->height = vnFontHeight;
}

static void SelectFile(Widget w, DfFileList *list)
{
  int n;
  DfFile *f;
  int i;
  XRectangle rc;

  n = list->nCur;

  if(list->nFiles <= n){
    return;
  }

  f = &list->files[n];

  if(f->nSelect){
    n = f->nSelect;
    f->nSelect = 0;
    f = list->files;
    for(i = 0; i < list->nFiles; i++){
      if(n < f->nSelect){
	f->nSelect--;
      }
      f++;
    }
    list->nSelects--;
  }else{
    list->nSelects++;
    f->nSelect = list->nSelects;
  }
  CalcCursorRect(list, n - list->nStart, &rc);
  XClearArea(XtDisplay(w), XtWindow(w), rc.x, rc.y, rc.width, rc.height, True);
}

static int GotoParent(Widget w, DfFileList *list)
{
  char *p;

  p = FindLastChar(list->pszCurDir, '/');
  if(!p){
    return 0;
  }

  if(p[1] == '\0' && p == list->pszCurDir){
    /*  /  */
    return 0;
  }

  if(p[1] == '\0'){
    p[0] = '\0';
    p = FindLastChar(list->pszCurDir, '/');
  }
  p[1] = '\0';
  return GetFiles(list, NULL);
}

