#include <stdio.h>
#include <strings.h>
#include <malloc.h>

#include <X11/Intrinsic.h>

#include "dftype.h"
#include "dfxfunc.h"
#define BUF_SIZE (4096)

static int extend(DfStr *s, int len);


int Str_Init(DfStr *s)
{
  s->str = bMalloc(BUF_SIZE);
  s->cnt = 1;
  s->len = 0;
  s->alloced_len = BUF_SIZE;

  return 0;
}

int Str_InitNull(DfStr *s)
{
  s->str = NULL;
  s->cnt = 1;
  s->len = 0;
  s->alloced_len = 0;

  return 0;
}


int Str_InitStr(DfStr *s, const char *str)
{
  int len;
  int a_len;

  if(!str){
    return Str_Init(s);
  }
  len = strlen(str);
  a_len = ((len + BUF_SIZE) / BUF_SIZE) * BUF_SIZE;

  s->str = bMalloc(a_len);
  strcpy(s->str, str);
  s->cnt = 1;
  s->len = len;
  s->alloced_len = a_len;

  return 0;
}

int Str_InitPsz(DfStr *s, const char *str)
{
  int len;

  len = str ? strlen(str) : 0;
  s->str = (char*)str;
  s->len = len;
  s->cnt = 1;
  s->alloced_len = len;

  return 0;
}


int Str_Free(DfStr *s, int fFree)
{
  s->cnt--;
  if(s->cnt){
    return 1;
  }

  bFree(s->str);
  if(fFree){
    bFree(s);
  }else{
    s->str = NULL;
    s->len = 0;
    s->alloced_len = 0;
  }
  return 1;
}

void Str_SetLength(DfStr *s, int len)
{
  s->str[len] = '\0';
  s->len = len;
}


char *Str_Get(DfStr *s)
{
  return s->str;
}

int Str_Length(DfStr *s)
{
  return s->len;
}

int Str_Add(DfStr *s, const char *str)
{
  int add;

  add = strlen(str);

  if(extend(s, add) == 0){
    return 0;
  }

  strcpy(s->str + s->len, str);
  s->len = s->len + add;

  if(s->str == NULL || s->str[0] == '\0'){
    fprintf(stderr, "what?\n");
  }
  return 1;
}

int Str_AddNUL(DfStr *s, const char *str)
{
  int add;

  add = strlen(str) + 1;

  if(extend(s, add) == 0){
    return 0;
  }

  strcpy(s->str + s->len, str);
  s->str[s->len + add - 1] = '\0';
  s->len = s->len + add;

  if(s->str == NULL || s->str[0] == '\0'){
    fprintf(stderr, "what?\n");
  }
  return 1;
}


int Str_AddChar(DfStr *s, char ch)
{
  if(extend(s, 1) == 0){
    return 0;
  }

  s->str[s->len] = ch;
  s->len++;
  s->str[s->len] = '\0';

  return 1;
}

int Str_Overwrite(DfStr *s, int idx, const char *str)
{
  int len;

  len = strlen(str);

  if(s->alloced_len < idx + len){
    if(extend(s, idx + len - s->alloced_len) == 0){
      return 0;
    }
  }

  strcpy(s->str + idx, str);
  s->len = idx + len;

  return 1;
}

int Str_Shrink(DfStr *s)
{
  char *p;

  p = bReAlloc(s->str, s->len + 1);
  if(p){
    s->str = p;
    s->alloced_len = s->len + 1;
  }
  return 1;
}


int Str_RegPath(DfStr *path)
{
  char *p;
  char *s;
  int len;

  p = path->str;
  s = p;

  while(*p){
    switch(*p){
    case '/':
      switch(p[1]){
      case '.':
        switch(p[2]){
        case '/':
        case '\0':
          p = p + 2;
          continue;
          break;
        case '.':
          if(p[3] == '/' || p[3] == '\0'){
            p = p + 3;
            /* back */
            while(path->str < s){
              s--;
              if(s[0] == '/'){
                break;
              }
            }
/*            if(p[0] == '\0'){
              s++;
            }*/
            continue;
          }
          break;
        }
        break;
      case '/':
        p = p + 2;
        while(p[1] == '/')
            p++;
        break;
      }
      break;
    }
    *s = *p;
    p++;
    s++;
  }

  len = s - path->str;
  if(len == 0){
    len = 1;
  }
  Str_SetLength(path, len);

  return p != s;
}

static int extend(DfStr *s, int len)
{
  char *p;

  len = s->len + len;
  if(s->alloced_len < len){
    /* extend */
    len += BUF_SIZE;
    len = (len / BUF_SIZE) * BUF_SIZE;
    p = bReAlloc(s->str, len);
    if(p == NULL){
#if MEMDEBUG
      fprintf(stderr, "no enough memory.\n");
#endif
      return 0;
    }
    s->str = p;
    s->alloced_len = len;
  }
  return 1;
}
